@@ -46,19 +46,22 @@
 #include <osmocom/ctrl/control_if.h>
 #include <osmocom/gsm/apn.h>
 
+#include <osmocom/gtp/pdp.h>
+#include <osmocom/gtp/gtp.h>
+
 #include "../lib/tun.h"
 #include "../lib/ippool.h"
 #include "../lib/syserr.h"
 #include "../lib/in46_addr.h"
 #include "../lib/gtp-kernel.h"
 #include "../lib/util.h"
-#include "../gtp/pdp.h"
-#include "../gtp/gtp.h"
 #include "../lib/icmpv6.h"
 #include "pco.h"
 #include "ggsn.h"
+#include "../gtp/gtp_internal.h"
+
+LLIST_HEAD(g_ggsn_list);
 
-static int ggsn_tun_fd_cb(struct osmo_fd *fd, unsigned int what);
 static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len);
 
 void ggsn_close_one_pdp(struct pdp_t *pdp)
@@ -92,6 +95,43 @@ static void pool_close_all_pdp(struct ippool_t *pool)
 	}
 }
 
+static struct apn_ctx *apn_alloc(struct ggsn_ctx *ggsn, const char *name)
+{
+	struct apn_ctx *apn;
+
+	apn = talloc_zero(ggsn, struct apn_ctx);
+	OSMO_ASSERT(apn);
+
+	apn->ggsn = ggsn;
+	apn->cfg.name = talloc_strdup(apn, name);
+	apn->cfg.shutdown = true;
+	apn->cfg.tx_gpdu_seq = true;
+	apn->cfg.mtu = MAX_DESIRED_APN_MTU;
+	INIT_LLIST_HEAD(&apn->cfg.name_list);
+
+	llist_add_tail(&apn->list, &ggsn->apn_list);
+	return apn;
+}
+
+struct apn_ctx *ggsn_find_apn(struct ggsn_ctx *ggsn, const char *name)
+{
+	struct apn_ctx *apn;
+
+	llist_for_each_entry(apn, &ggsn->apn_list, list) {
+		if (!strcmp(apn->cfg.name, name))
+			return apn;
+	}
+	return NULL;
+}
+
+struct apn_ctx *ggsn_find_or_create_apn(struct ggsn_ctx *ggsn, const char *name)
+{
+	struct apn_ctx *apn = ggsn_find_apn(ggsn, name);
+	if (!apn)
+		apn = apn_alloc(ggsn, name);
+	return apn;
+}
+
 int apn_stop(struct apn_ctx *apn)
 {
 	LOGPAPN(LOGL_NOTICE, apn, "Stopping\n");
@@ -109,7 +149,6 @@ int apn_stop(struct apn_ctx *apn)
 		if (apn->cfg.gtpu_mode == APN_GTPU_MODE_TUN) {
 			/* release tun device */
 			LOGPAPN(LOGL_INFO, apn, "Closing TUN device %s\n", apn->tun.tun->devname);
-			osmo_fd_unregister(&apn->tun.fd);
 		}
 		tun_free(apn->tun.tun);
 		apn->tun.tun = NULL;
@@ -187,26 +226,18 @@ int apn_start(struct apn_ctx *apn)
 	switch (apn->cfg.gtpu_mode) {
 	case APN_GTPU_MODE_TUN:
 		LOGPAPN(LOGL_INFO, apn, "Opening TUN device %s\n", apn->tun.cfg.dev_name);
-		if (tun_new(&apn->tun.tun, apn->tun.cfg.dev_name, false, -1, -1)) {
+		apn->tun.tun = tun_alloc_tundev(apn->tun.cfg.dev_name);
+		if (!apn->tun.tun) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to configure tun device\n");
 			return -1;
 		}
 		LOGPAPN(LOGL_INFO, apn, "Opened TUN device %s\n", apn->tun.tun->devname);
 
-		/* Register with libosmcoore */
-		osmo_fd_setup(&apn->tun.fd, apn->tun.tun->fd, OSMO_FD_READ, ggsn_tun_fd_cb, apn, 0);
-		osmo_fd_register(&apn->tun.fd);
-
 		/* Set TUN library callback */
 		tun_set_cb_ind(apn->tun.tun, cb_tun_ind);
 		break;
 	case APN_GTPU_MODE_KERNEL_GTP:
 		LOGPAPN(LOGL_INFO, apn, "Opening Kernel GTP device %s\n", apn->tun.cfg.dev_name);
-		if (apn->cfg.apn_type_mask & (APN_TYPE_IPv6|APN_TYPE_IPv4v6)) {
-			LOGPAPN(LOGL_ERROR, apn, "Kernel GTP currently supports only IPv4\n");
-			apn_stop(apn);
-			return -1;
-		}
 		if (gsn == NULL) {
 			/* skip bringing up the APN now if the GSN is not initialized yet.
 			 * This happens during initial load of the config file, as the
@@ -216,7 +247,8 @@ int apn_start(struct apn_ctx *apn)
 			return 0;
 		}
 		/* use GTP kernel module for data packet encapsulation */
-		if (tun_new(&apn->tun.tun, apn->tun.cfg.dev_name, true, gsn->fd0, gsn->fd1u)) {
+		apn->tun.tun = tun_alloc_gtpdev(apn->tun.cfg.dev_name, gsn->fd0, gsn->fd1u);
+		if (!apn->tun.tun) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to configure Kernel GTP device\n");
 			return -1;
 		}
@@ -234,7 +266,7 @@ int apn_start(struct apn_ctx *apn)
 	if (apn->v4.cfg.ifconfig_prefix.addr.len) {
 		LOGPAPN(LOGL_INFO, apn, "Setting tun IP address %s\n",
 			in46p_ntoa(&apn->v4.cfg.ifconfig_prefix));
-		if (tun_addaddr(apn->tun.tun, &apn->v4.cfg.ifconfig_prefix.addr, NULL,
+		if (tun_addaddr(apn->tun.tun, &apn->v4.cfg.ifconfig_prefix.addr,
 				apn->v4.cfg.ifconfig_prefix.prefixlen)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv4 address %s: %s\n",
 				in46p_ntoa(&apn->v4.cfg.ifconfig_prefix), strerror(errno));
@@ -246,7 +278,7 @@ int apn_start(struct apn_ctx *apn)
 	if (apn->v6.cfg.ifconfig_prefix.addr.len) {
 		LOGPAPN(LOGL_INFO, apn, "Setting tun IPv6 address %s\n",
 			in46p_ntoa(&apn->v6.cfg.ifconfig_prefix));
-		if (tun_addaddr(apn->tun.tun, &apn->v6.cfg.ifconfig_prefix.addr, NULL,
+		if (tun_addaddr(apn->tun.tun, &apn->v6.cfg.ifconfig_prefix.addr,
 				apn->v6.cfg.ifconfig_prefix.prefixlen)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv6 address %s: %s. "
 				"Ensure you have ipv6 support and not used the disable_ipv6 sysctl?\n",
@@ -259,7 +291,7 @@ int apn_start(struct apn_ctx *apn)
 	if (apn->v6.cfg.ll_prefix.addr.len) {
 		LOGPAPN(LOGL_INFO, apn, "Setting tun IPv6 link-local address %s\n",
 			in46p_ntoa(&apn->v6.cfg.ll_prefix));
-		if (tun_addaddr(apn->tun.tun, &apn->v6.cfg.ll_prefix.addr, NULL,
+		if (tun_addaddr(apn->tun.tun, &apn->v6.cfg.ll_prefix.addr,
 				apn->v6.cfg.ll_prefix.prefixlen)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv6 link-local address %s: %s. "
 				"Ensure you have ipv6 support and not used the disable_ipv6 sysctl?\n",
@@ -270,6 +302,23 @@ int apn_start(struct apn_ctx *apn)
 		apn->v6_lladdr = apn->v6.cfg.ll_prefix.addr.v6;
 	}
 
+	rc = osmo_netdev_ifupdown(apn->tun.tun->netdev, true);
+	if (rc < 0) {
+		LOGPAPN(LOGL_ERROR, apn, "Failed to set tun interface UP: %s\n", strerror(errno));
+		apn_stop(apn);
+		return -1;
+	}
+
+	if (apn->tun.cfg.mtu_apply) {
+		rc = osmo_netdev_set_mtu(apn->tun.tun->netdev, apn->cfg.mtu);
+		if (rc < 0) {
+			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun interface MTU %u: %s\n",
+				apn->cfg.mtu, strerror(errno));
+			apn_stop(apn);
+			return -1;
+		}
+	}
+
 	if (apn->tun.cfg.ipup_script) {
 		LOGPAPN(LOGL_INFO, apn, "Running ip-up script %s\n",
 			apn->tun.cfg.ipup_script);
@@ -428,7 +477,7 @@ static struct sgsn_peer* ggsn_find_or_create_sgsn(struct ggsn_ctx *ggsn, struct
 	return sgsn;
 }
 
-int create_context_ind(struct pdp_t *pdp)
+static int create_context_ind(struct pdp_t *pdp)
 {
 	static char name_buf[256];
 	struct gsn_t *gsn = pdp->gsn;
@@ -536,10 +585,11 @@ int create_context_ind(struct pdp_t *pdp)
 
 	in46a_to_eua(addr, num_addr, &pdp->eua);
 
-	if (apn->cfg.gtpu_mode == APN_GTPU_MODE_KERNEL_GTP && apn_supports_ipv4(apn)) {
-		/* TODO: In IPv6, EUA doesn't contain the actual IP addr/prefix! */
+	if (apn->cfg.gtpu_mode == APN_GTPU_MODE_KERNEL_GTP) {
 		if (gtp_kernel_tunnel_add(pdp, apn->tun.cfg.dev_name) < 0) {
 			LOGPPDP(LOGL_ERROR, pdp, "Cannot add tunnel to kernel: %s\n", strerror(errno));
+			if (addrv6 && errno == EINVAL)
+				LOGPPDP(LOGL_ERROR, pdp, "Maybe your kernel does not support GTP-U with IPv6 yet?\n");
 			gtp_create_context_resp(gsn, pdp, GTPCAUSE_SYS_FAIL);
 			return 0;
 		}
@@ -584,7 +634,43 @@ err_wrong_af:
 	return 0;
 }
 
-/* Internet-originated IP packet, needs to be sent via GTP towards MS */
+static int update_context_ind(struct pdp_t *pdp)
+{
+	char apn_name[256];
+	struct gsn_t *gsn = pdp->gsn;
+	struct ggsn_ctx *ggsn = gsn->priv;
+	struct apn_ctx *apn;
+	bool apn_found = false;
+	int rc;
+
+	if (!osmo_apn_to_str(apn_name, pdp->apn_use.v, pdp->apn_use.l)) {
+		LOGPPDP(LOGL_ERROR, pdp, "Unable to decode associated APN len=%d buf: %s\n",
+			pdp->apn_use.l, osmo_hexdump(pdp->apn_use.v, pdp->apn_use.l));
+		return gtp_update_context_resp(ggsn->gsn, pdp, GTPCAUSE_MISSING_APN);
+	}
+
+	llist_for_each_entry (apn, &ggsn->apn_list, list) {
+		if (strncmp(apn_name, apn->cfg.name, sizeof(apn_name)) != 0)
+			continue;
+		apn_found = true;
+		break;
+	}
+	if (!apn_found) {
+		LOGPPDP(LOGL_ERROR, pdp, "Unable to find associated APN %s\n", apn_name);
+		return gtp_update_context_resp(ggsn->gsn, pdp, GTPCAUSE_MISSING_APN);
+	}
+
+	if (apn->cfg.gtpu_mode == APN_GTPU_MODE_KERNEL_GTP) {
+		/* Update the kernel with the potentially new remote data IP address + TEID */
+		gtp_kernel_tunnel_del(pdp, apn->tun.cfg.dev_name);
+		gtp_kernel_tunnel_add(pdp, apn->tun.cfg.dev_name);
+	}
+	rc = gtp_update_context_resp(ggsn->gsn, pdp, GTPCAUSE_ACC_REQ);
+	return rc;
+}
+
+/* Rx Internet-originated IP packet from our tun iface, needs to be sent via
+ * GTPU towards MS. */
 static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 {
 	struct apn_ctx *apn = tun->priv;
@@ -664,8 +750,10 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 	return 0;
 }
 
-/* MS-originated GTP1-U packet, needs to be sent via TUN device */
-static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
+/* Rx MS-originated GTP-U packet, needs to be sent via TUN device.
+ * "pack" contains the GTP-U payload once decapsulated from GTP-U by libgtp.
+ * Hence, "pack" should contain an IP packet. */
+static int cb_gtpu_data_ind(struct pdp_t *pdp, void *pack, unsigned len)
 {
 	struct iphdr *iph = (struct iphdr *)pack;
 	struct ip6_hdr *ip6h = (struct ip6_hdr *)pack;
@@ -699,10 +787,15 @@ static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 			return -1;
 		}
 
-		/* daddr: all-routers multicast addr */
-		if (IN6_ARE_ADDR_EQUAL(&ip6h->ip6_dst, &all_router_mcast_addr))
-			return handle_router_mcast(pdp->gsn, pdp, &peer->addr.v6,
-						&apn->v6_lladdr, pack, len);
+		if (IN6_IS_ADDR_MULTICAST(&ip6h->ip6_dst)) {
+			/* daddr: all-routers multicast addr */
+			if (IN6_ARE_ADDR_EQUAL(&ip6h->ip6_dst, &all_router_mcast_addr))
+				return handle_router_mcast(pdp->gsn, pdp, &peer->addr.v6,
+							   &apn->v6_lladdr, apn->cfg.mtu, pack, len);
+			/* daddr: solicited-node multicast addr */
+			if (memcmp(&ip6h->ip6_dst.s6_addr, solicited_node_mcast_addr_prefix, sizeof(solicited_node_mcast_addr_prefix)) == 0)
+				return handle_solicited_node_mcast(pack, len);
+		}
 		break;
 	case 4:
 		peer = pdp_get_peer_ipv(pdp, false);
@@ -724,17 +817,7 @@ static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 			osmo_hexdump(pack, len));
 		return -1;
 	}
-	return tun_encaps((struct tun_t *)pdp->ipif, pack, len);
-}
-
-/* callback for tun device osmocom select loop integration */
-static int ggsn_tun_fd_cb(struct osmo_fd *fd, unsigned int what)
-{
-	struct apn_ctx *apn = fd->data;
-
-	OSMO_ASSERT(what & OSMO_FD_READ);
-
-	return tun_decaps(apn->tun.tun);
+	return tun_inject_pkt((struct tun_t *)pdp->ipif, pack, len);
 }
 
 /* callback for libgtp osmocom select loop integration */
@@ -810,6 +893,42 @@ static int cb_recovery3(struct gsn_t *gsn, struct sockaddr_in *peer, struct pdp_
 	return sgsn_peer_handle_recovery(sgsn, pdp, recovery);
 }
 
+static struct ggsn_ctx *ggsn_alloc(void *ctx, const char *name)
+{
+	struct ggsn_ctx *ggsn;
+
+	ggsn = talloc_zero(ctx, struct ggsn_ctx);
+	OSMO_ASSERT(ggsn);
+
+	ggsn->cfg.name = talloc_strdup(ggsn, name);
+	ggsn->cfg.state_dir = talloc_strdup(ggsn, "/tmp");
+	ggsn->cfg.shutdown = true;
+	INIT_LLIST_HEAD(&ggsn->apn_list);
+	INIT_LLIST_HEAD(&ggsn->sgsn_list);
+
+	llist_add_tail(&ggsn->list, &g_ggsn_list);
+	return ggsn;
+}
+
+struct ggsn_ctx *ggsn_find(const char *name)
+{
+	struct ggsn_ctx *ggsn;
+
+	llist_for_each_entry(ggsn, &g_ggsn_list, list) {
+		if (!strcmp(ggsn->cfg.name, name))
+			return ggsn;
+	}
+	return NULL;
+}
+
+struct ggsn_ctx *ggsn_find_or_create(void *ctx, const char *name)
+{
+	struct ggsn_ctx *ggsn = ggsn_find(name);
+	if (!ggsn)
+		ggsn = ggsn_alloc(ctx, name);
+	return ggsn;
+}
+
 /* Start a given GGSN */
 int ggsn_start(struct ggsn_ctx *ggsn)
 {
@@ -849,9 +968,10 @@ int ggsn_start(struct ggsn_ctx *ggsn)
 	rc = osmo_fd_register(&ggsn->gtp_fd1u);
 	OSMO_ASSERT(rc == 0);
 
-	gtp_set_cb_data_ind(ggsn->gsn, encaps_tun);
-	gtp_set_cb_delete_context(ggsn->gsn, delete_context);
+	gtp_set_cb_data_ind(ggsn->gsn, cb_gtpu_data_ind);
 	gtp_set_cb_create_context_ind(ggsn->gsn, create_context_ind);
+	gtp_set_cb_update_context_ind(ggsn->gsn, update_context_ind);
+	gtp_set_cb_delete_context(ggsn->gsn, delete_context);
 	gtp_set_cb_conf(ggsn->gsn, cb_conf);
 	gtp_set_cb_recovery3(ggsn->gsn, cb_recovery3);
 
