@@ -50,76 +50,11 @@
 #define IFCONFIG_STR	"GGSN-based interface configuration\n"
 #define GGSN_STR	"Gateway GPRS Support NODE (GGSN)\n"
 
-LLIST_HEAD(g_ggsn_list);
-
 enum ggsn_vty_node {
 	GGSN_NODE = _LAST_OSMOVTY_NODE + 1,
 	APN_NODE,
 };
 
-struct ggsn_ctx *ggsn_find(const char *name)
-{
-	struct ggsn_ctx *ggsn;
-
-	llist_for_each_entry(ggsn, &g_ggsn_list, list) {
-		if (!strcmp(ggsn->cfg.name, name))
-			return ggsn;
-	}
-	return NULL;
-}
-
-struct ggsn_ctx *ggsn_find_or_create(void *ctx, const char *name)
-{
-	struct ggsn_ctx *ggsn;
-
-	ggsn = ggsn_find(name);
-	if (ggsn)
-		return ggsn;
-
-	ggsn = talloc_zero(ctx, struct ggsn_ctx);
-	if (!ggsn)
-		return NULL;
-
-	ggsn->cfg.name = talloc_strdup(ggsn, name);
-	ggsn->cfg.state_dir = talloc_strdup(ggsn, "/tmp");
-	ggsn->cfg.shutdown = true;
-	INIT_LLIST_HEAD(&ggsn->apn_list);
-	INIT_LLIST_HEAD(&ggsn->sgsn_list);
-
-	llist_add_tail(&ggsn->list, &g_ggsn_list);
-	return ggsn;
-}
-
-struct apn_ctx *ggsn_find_apn(struct ggsn_ctx *ggsn, const char *name)
-{
-	struct apn_ctx *apn;
-
-	llist_for_each_entry(apn, &ggsn->apn_list, list) {
-		if (!strcmp(apn->cfg.name, name))
-			return apn;
-	}
-	return NULL;
-}
-
-struct apn_ctx *ggsn_find_or_create_apn(struct ggsn_ctx *ggsn, const char *name)
-{
-	struct apn_ctx *apn = ggsn_find_apn(ggsn, name);
-	if (apn)
-		return apn;
-
-	apn = talloc_zero(ggsn, struct apn_ctx);
-	if (!apn)
-		return NULL;
-	apn->ggsn = ggsn;
-	apn->cfg.name = talloc_strdup(apn, name);
-	apn->cfg.shutdown = true;
-	apn->cfg.tx_gpdu_seq = true;
-	INIT_LLIST_HEAD(&apn->cfg.name_list);
-
-	llist_add_tail(&apn->list, &ggsn->apn_list);
-	return apn;
-}
-
 /* GGSN Node */
 
 static struct cmd_node ggsn_node = {
@@ -477,13 +412,46 @@ DEFUN(cfg_apn_gtpu_mode, cfg_apn_gtpu_mode_cmd,
 DEFUN(cfg_apn_tun_dev_name, cfg_apn_tun_dev_name_cmd,
 	"tun-device NAME",
 	"Configure tun device name\n"
-	"TUN device name")
+	"TUN device name\n")
 {
 	struct apn_ctx *apn = (struct apn_ctx *) vty->index;
 	osmo_talloc_replace_string(apn, &apn->tun.cfg.dev_name, argv[0]);
 	return CMD_SUCCESS;
 }
 
+/* MAX_POSSIBLE_APN_MTU = 9000
+ * MAX_DESIRED_APN_MTU = 1420 */
+DEFUN(cfg_apn_mtu, cfg_apn_mtu_cmd,
+	"mtu (<0-" OSMO_STRINGIFY_VAL(MAX_POSSIBLE_APN_MTU) ">|default) [apply]",
+	"Configure announced MTU\n"
+	"MTU of the APN, announced to the UE\n"
+	"Default value of the MTU of the APN (1420)\n"
+	"Apply the MTU on the tun-device of the APN\n")
+{
+	struct apn_ctx *apn = (struct apn_ctx *) vty->index;
+	int rc;
+
+	if (strcmp(argv[0], "default") == 0)
+		apn->cfg.mtu = MAX_DESIRED_APN_MTU;
+	else
+		apn->cfg.mtu = atoi(argv[0]);
+	apn->tun.cfg.mtu_apply = (argc > 1);
+
+	if (apn->tun.cfg.mtu_apply && apn->tun.tun) {
+		rc = osmo_netdev_set_mtu(apn->tun.tun->netdev, apn->cfg.mtu);
+		if (rc < 0) {
+			char buf_err[128];
+			strerror_r(errno, buf_err, sizeof(buf_err));
+			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun interface MTU %u: %s (%d)\n",
+				apn->cfg.mtu, buf_err, rc);
+			vty_out(vty, "%% Failed to set tun interface MTU %u: %s (%d)%s",
+				apn->cfg.mtu, buf_err, rc, VTY_NEWLINE);
+			return CMD_WARNING;
+		}
+	}
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_apn_ipup_script, cfg_apn_ipup_script_cmd,
 	"ipup-script PATH",
 	"Configure name/path of ip-up script\n"
@@ -754,6 +722,9 @@ static void config_write_apn(struct vty *vty, struct apn_ctx *apn)
 			VTY_NEWLINE);
 	}
 
+	vty_out(vty, "  mtu %" PRIu16 "%s%s", apn->cfg.mtu,
+		apn->tun.cfg.mtu_apply ? " apply" : "", VTY_NEWLINE);
+
 	if (!apn->cfg.tx_gpdu_seq)
 		vty_out(vty, "  no g-pdu tx-sequence-numbers%s", VTY_NEWLINE);
 
@@ -1131,6 +1102,7 @@ int ggsn_vty_init(void)
 	install_element(APN_NODE, &cfg_apn_type_support_cmd);
 	install_element(APN_NODE, &cfg_apn_no_type_support_cmd);
 	install_element(APN_NODE, &cfg_apn_tun_dev_name_cmd);
+	install_element(APN_NODE, &cfg_apn_mtu_cmd);
 	install_element(APN_NODE, &cfg_apn_ipup_script_cmd);
 	install_element(APN_NODE, &cfg_apn_no_ipup_script_cmd);
 	install_element(APN_NODE, &cfg_apn_ipdown_script_cmd);
