@@ -23,6 +23,7 @@
 
 #include <osmocom/core/application.h>
 #include <osmocom/core/msgb.h>
+#include <osmocom/core/select.h>
 
 #include <ctype.h>
 #include <netdb.h>
@@ -1195,7 +1196,7 @@ static int ping_finish()
 	return 0;
 }
 
-static int encaps_ping4(struct pdp_t *pdp, void *pack, unsigned len)
+static int cb_gtpu_data_ind_ping4(struct pdp_t *pdp, void *pack, unsigned len)
 {
 	struct timeval tv;
 	struct timeval *tp;
@@ -1257,7 +1258,7 @@ static int encaps_ping4(struct pdp_t *pdp, void *pack, unsigned len)
 	return 0;
 }
 
-static int encaps_ping6(struct pdp_t *pdp, struct ip6_hdr *ip6h, unsigned len)
+static int cb_gtpu_data_ind_ping6(struct pdp_t *pdp, struct ip6_hdr *ip6h, unsigned len)
 {
 	const struct icmpv6_echo_hdr *ic6h = (struct icmpv6_echo_hdr *) ((uint8_t*)ip6h + sizeof(*ip6h));
 	struct timeval tv;
@@ -1324,7 +1325,7 @@ static int encaps_ping6(struct pdp_t *pdp, struct ip6_hdr *ip6h, unsigned len)
 }
 
 /* Handle a received ping packet. Print out line and update statistics. */
-static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
+static int cb_gtpu_data_ind_ping(struct pdp_t *pdp, void *pack, unsigned len)
 {
 	struct iphdr *iph = (struct iphdr *)pack;
 	struct timeval tv;
@@ -1342,9 +1343,9 @@ static int encaps_ping(struct pdp_t *pdp, void *pack, unsigned len)
 	}
 	switch(iph->version) {
 	case 4:
-		return encaps_ping4(pdp, pack, len);
+		return cb_gtpu_data_ind_ping4(pdp, pack, len);
 	case 6:
-		return encaps_ping6(pdp, (struct ip6_hdr *)pack, len);
+		return cb_gtpu_data_ind_ping6(pdp, (struct ip6_hdr *)pack, len);
 	default:
 		SYS_ERR(DSGSN, LOGL_ERROR, 0, "Unknown ip header version %d", iph->version);
 		return -1;
@@ -1648,7 +1649,7 @@ static int create_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
 			if (addr[i].len == 16)
 				prefixlen = 64;
 			/* printf("Setting up interface and routing\n"); */
-			tun_addaddr(tun, &addr[i], NULL, prefixlen);
+			tun_addaddr(tun, &addr[i], prefixlen);
 			if (options.defaultroute) {
 				if (in46a_is_v4(&addr[i])) {
 					struct in_addr rm;
@@ -1785,7 +1786,7 @@ static void handle_router_adv(struct pdp_t *pdp, struct ip6_hdr *ip6h, struct ic
 					}
 				}
 #endif
-				rc = tun_addaddr(tun, &addr, NULL, opt_prefix->prefix_len);
+				rc = tun_addaddr(tun, &addr, opt_prefix->prefix_len);
 				if (rc < 0) {
 					SYS_ERR(DSGSN, LOGL_ERROR, 0, "Failed to add addr %s to tun %s",
 						in46a_ntoa(&addr), tun->devname);
@@ -1811,7 +1812,7 @@ static void handle_router_adv(struct pdp_t *pdp, struct ip6_hdr *ip6h, struct ic
 	}
 }
 
-static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
+static int cb_gtpu_data_ind(struct pdp_t *pdp, void *pack, unsigned len)
 {
 	struct iphdr *iph = (struct iphdr *)pack;
 	struct icmpv6_radv_hdr *ra;
@@ -1825,8 +1826,8 @@ static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 	break;
 	}
 
-	/*  printf("encaps_tun. Packet received: forwarding to tun\n"); */
-	return tun_encaps((struct tun_t *)pdp->ipif, pack, len);
+	/*  printf("cb_gtpu_data_ind. Packet received: forwarding to tun\n"); */
+	return tun_inject_pkt((struct tun_t *)pdp->ipif, pack, len);
 }
 
 int main(int argc, char **argv)
@@ -1882,9 +1883,9 @@ int main(int argc, char **argv)
 	gtp_set_cb_delete_context(gsn, delete_context);
 	gtp_set_cb_conf(gsn, _gtp_cb_conf);
 	if (options.createif)
-		gtp_set_cb_data_ind(gsn, encaps_tun);
+		gtp_set_cb_data_ind(gsn, cb_gtpu_data_ind);
 	else
-		gtp_set_cb_data_ind(gsn, encaps_ping);
+		gtp_set_cb_data_ind(gsn, cb_gtpu_data_ind_ping);
 
 #if defined(__linux__)
 	if ((options.createif) && (options.netns)) {
@@ -1904,7 +1905,8 @@ int main(int argc, char **argv)
 	if (options.createif) {
 		printf("Setting up interface\n");
 		/* Create a tunnel interface */
-		if (tun_new((struct tun_t **)&tun, options.tun_dev_name, false, -1, -1)) {
+		tun = tun_alloc_tundev(options.tun_dev_name);
+		if (!tun) {
 			SYS_ERR(DSGSN, LOGL_ERROR, 0,
 				"Failed to create tun");
 			exit(1);
@@ -1924,8 +1926,8 @@ int main(int argc, char **argv)
 #endif
 
 		tun_set_cb_ind(tun, cb_tun_ind);
-		if (tun->fd > maxfd)
-			maxfd = tun->fd;
+		if (tun->tundev.fd > maxfd)
+			maxfd = tun->tundev.fd;
 
 		if (proc_ipv6_conf_write(tun->devname, "accept_ra", "0") < 0) {
 			SYS_ERR(DSGSN, LOGL_ERROR, 0,
@@ -1935,7 +1937,7 @@ int main(int argc, char **argv)
 	}
 
 	if ((options.createif) && (options.netaddr.len)) {
-		tun_addaddr(tun, &options.netaddr, NULL, options.prefixlen);
+		tun_addaddr(tun, &options.netaddr, options.prefixlen);
 		if (options.defaultroute) {
 			if (in46a_is_v4(&options.netaddr)) {
 				struct in_addr rm;
@@ -2158,7 +2160,7 @@ int main(int argc, char **argv)
 
 		FD_ZERO(&fds);
 		if (tun)
-			FD_SET(tun->fd, &fds);
+			FD_SET(tun->tundev.fd, &fds);
 		FD_SET(gsn->fd0, &fds);
 		FD_SET(gsn->fd1c, &fds);
 		FD_SET(gsn->fd1u, &fds);
@@ -2186,9 +2188,8 @@ int main(int argc, char **argv)
 
 		if (!signal_received) {
 
-			if ((tun) && FD_ISSET(tun->fd, &fds) && tun_decaps(tun) < 0) {
-				SYS_ERR(DSGSN, LOGL_ERROR, 0,
-					"TUN decaps failed");
+			if ((tun) && FD_ISSET(tun->tundev.fd, &fds)) {
+				osmo_select_main(1);
 			}
 
 			if (FD_ISSET(gsn->fd0, &fds))
