@@ -10,8 +10,9 @@
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/bits.h>
 
+#include <osmocom/gtp/gtpie.h>
+
 #include "../../lib/syserr.h"
-#include "../../gtp/gtpie.h"
 
 static const uint8_t in[] = { 1,2,3,4,5,6 };
 static uint8_t buf[256];
@@ -113,7 +114,7 @@ int main(int argc, char **argv)
 	msgb_talloc_ctx_init(tall_ctx, 0);
 	osmo_init_logging2(tall_ctx, &log_info);
 	log_set_use_color(osmo_stderr_target, 0);
-	log_set_print_filename(osmo_stderr_target, 0);
+	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_NONE);
 
 	srand(time(NULL));
 
